"""
EnamineFunctions.py - functions for interacting with Enamine Web Services

Use of this code is subject to the terms of the StarDrop License Agreement and
use, copying or redistribution of this file for use by other persons or
organisations is prohibited

Copyright(C) Optibrium Ltd 2019
stardrop-support@optibrium.com
"""

import PythonQt
import EnamineQueryLogger
from EnamineWebServices import EnamineWebServices
import Enamine_credentials
import json
import os
import app
import codecs
import socket
import itertools


log = EnamineQueryLogger.EnamineQueryLogger('enaminequerytool.log')
socket.setdefaulttimeout(20)

cred = None
configured = False


def _getCredentials(appui):
    """Get a user name and password from the user and store in a CredentialsManager object for use during the lifetime
    of the session.
    """
    global cred
    if not cred or cred.configured == False:
        dlg = PythonQt.LogInDialog()
        dlg.setWindowTitle("Enamine Login")
        if dlg.run():
            try:
                username = dlg.getUser()
                password = dlg.getPassword()
                if username == "" or password == "":
                    appui.show(
                        "User and password must be completed")
                else:
                    cred = Enamine_credentials.CredentialsManager(username, password, configured)
            except ValueError as ve:
                print(ve)
            except Exception as e:
                print(e)

    else:

        credentials = vars(cred)
        username = credentials.get('_username')
        password = credentials.get('_password')

    if not username or not password:
        raise ValueError("Could not locate authentication details for Enamine "
                         "API - cannot create link to web services")

    return username, password


def _create_enamine_web_services(appui):
    """
    Create an instance of the EnamineWebServices object either based on user
    input for details or from the session credentials.

    If the credentials have not been confirmed or do not exist then the user will be
    prompted to provide new details.

    This function may throw ValueError if after all this logic it still has no
    username or password.

    :return: an instance of EnamineWebServices
    :rtype: EnamineWebServices
    """
    username, password = _getCredentials(appui)

    if not username or not password:
        raise ValueError("Could not locate authentication details for Enamine "
                         "API - cannot create link to web services")

    web_services = EnamineWebServices(username, password)
    return web_services, username, password

def _create_string_entry(value):
    """

    :param value:   The string value to store in the new string entry
    :return:    The new string data entry
    """
    return app.Dataset.string(str(value))


def _create_columns(column_definitions):
    """
    Create columns for a dataset based on a set of column definitions. This
    is a convenience function that allows us to cut down on lots of repeated
    code to make columns.

    :param column_definitions:  The list of columns to create. Each column
                                should consist of a name, a type and None or a
                                list of category names depending on type
    :return: A tuple containing column headers and category dictionaries
    """
    cols = list()

    for column_definition in column_definitions:
        header, col_type, categories = column_definition

        if col_type == app.mol:
            cols.append(app.Dataset.meta(header, app.mol))
        elif col_type == app.cont:
            cols.append(app.Dataset.createNumberHeader(header))
        elif col_type == app.string:
            cols.append(app.Dataset.meta(header))

    return cols

def find_threshold(appui):
    """
    Function for requesting a similarity threshold percentage from the user.

    :param appui:       The stardrop appui object
    :return threshold:  The similarity threshold input by the user as a float
    """
    threshold = (appui.query('Similarity',
                             'Enter a similarity value:\n(Note: This must be 0.4 or higher. If the value you choose is low, you may\nsee errors because too many results are returned)',
                             '0.9'))

    threshold = float(threshold)

    # The threshold needs to be 0.4 or above, otherwise the API won't return results.
    if threshold < 0.40 or threshold > 1:
        threshold = find_threshold(appui)

    return(threshold)


def _unpack_packs(dataset, packs):
    """
    Find the 1 g/mg amount and price in packs and package that into the dataset. If there aren't any packs, a ?
    is added. If there isn't a 1 g/mg amount, use the first amount and price. This function is called when the
    user selects to show one amount and price.

    :param dataset:         The dataset to write new rows to
    :param packs:           A list of dictionaries with amounts that are available and their prices
    :return: None
    """
    # Where there is more than one amount/price, add only 1.0 g or mg. Check is used to make sure an amount and price
    # is added to the dataset as there could be instances where there isn't a 1 g/mg price.
    check = False
    if len(packs) > 1:
        for pack_ in packs:
            if float(pack_.get("amount")) == 1.0:
                if pack_.get("measure") == "g":
                    dataset.cols[7].append(
                        app.Dataset.cont(float(pack_.get("amount")), 0.0))
                else:
                    dataset.cols[7].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
                if pack_.get("measure") == "mg":
                    dataset.cols[8].append(
                        app.Dataset.cont(float(pack_.get("amount")), 0.0))
                else:
                    dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

                dataset.cols[9].append(
                    app.Dataset.cont(pack_.get("price"), 0.0))
                check = True
            else:
                continue

        if check == False:
            if pack_.get("measure") == "g":
                dataset.cols[7].append(
                    app.Dataset.cont(float(packs[0].get("amount")), 0.0))
            else:
                dataset.cols[7].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
            if pack_.get("measure") == "mg":
                dataset.cols[8].append(
                    app.Dataset.cont(float(pack_.get("amount")), 0.0))
            else:
                dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

            dataset.cols[9].append(
                app.Dataset.cont(packs[0].get("price"), 0.0))

    elif len(packs) == 1:
        if packs[0].get("measure") == "g":
            dataset.cols[7].append(
                app.Dataset.cont(float(packs[0].get("amount")), 0.0))
        else:
            dataset.cols[7].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
        if packs[0].get("measure") == "mg":
            dataset.cols[8].append(
                app.Dataset.cont(float(pack_.get("amount")), 0.0))
        else:
            dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

        dataset.cols[9].append(
            app.Dataset.cont(packs[0].get("price"), 0.0))

    else:
        dataset.cols[7].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
        dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
        dataset.cols[9].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

    return dataset

def _unpack_packs_ID_search(dataset, packs):
    """
    Find the 1 g/mg amount and price in packs and package that into the dataset. If there aren't any packs, a
    ? is added. If there isn't a 1 g/mg amount, use the first amount and price. This function is called when
    the user selects to show one amount and price.

    :param dataset:         The dataset to write new rows to
    :param packs:           A list of dictionaries with amounts that are available and their prices
    :return: None
    """
    check = False

    if len(packs) > 1:
        for pack_ in packs:
            if float(pack_.get("amount")) == 1.0:
                dataset.cols[7].append(
                    _create_string_entry(str(pack_.get("amount")) + " " + str(pack_.get("measure"))))
                dataset.cols[8].append(
                    app.Dataset.cont(pack_.get("price"), 0.0))
                check = True
            else:
                continue

        if check == False:
            dataset.cols[7].append(
                _create_string_entry(str(packs[0].get("amount")) + " " + str(pack_.get("measure"))))
            dataset.cols[8].append(
                app.Dataset.cont(packs[0].get("price"), 0.0))

    elif len(packs) == 1:
        dataset.cols[7].append(
            _create_string_entry(str(packs[0].get("amount")) + " " + str(packs[0].get("measure"))))
        dataset.cols[8].append(
            app.Dataset.cont(packs[0].get("price"), 0.0))

    else:
        dataset.cols[7].append(
            _create_string_entry(""))
        dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

    return dataset

def sketch_structure_query(appui):
    """
        Perform a structural search query based on the provided structure and using
        the specified search type (and similarity range if specified).

        :param appui:    The stardrop appui object
        :return: None
        """
    try:
        web_services, username, password = _create_enamine_web_services(appui)

        dlg = PythonQt.StructureQueryDialog()
        dlg.setPossibleMatchTypes(["MATCH", "SIMILARITY", "EQUALS"])
        if dlg.run():
            query = dlg.selectedQuery()
            smiles, operator, similarity = query
            log.log("running Structure query %s" % smiles)
            print("Running structure query for " + smiles)

        dataset = run_structure_query(appui, web_services, smiles, operator, similarity, username, password)

        if dataset:
            try:
                packdata = dataset.pack()
            except Exception as e:
                print("type error: " + str(e))
            appui.exportDS("Structure Query Result", packdata)


    except Exception as e:
        log.log(repr(e))


def structure_query_selected(appui):
    """
        Perform a structural search query using the selected structure and
        an input similarity value.

        :param appui:    The stardrop appui object
        :return: None
        """

    try:
        web_services, username, password = _create_enamine_web_services(appui)

        # get currently selected molecule
        SMILES_input = appui.getCurrentMolecules()

        # if no molecule is selected, tell user to select one and re-run
        if not SMILES_input:
            appui.show(
                "Please select a molecule before searching for similar compounds")
        elif len(SMILES_input) > 1:
            appui.show(
                "Please select exactly one molecule")

        else:
            SMILES_input = SMILES_input[0]
            print("Running structure query for " + SMILES_input)

            threshold = find_threshold(appui)
            threshold_value = str(threshold)

            operator = "LIKE"

            dataset = run_structure_query(appui, web_services, SMILES_input, operator, threshold_value, username, password)

            if dataset:
                try:
                    packdata = dataset.pack()
                except Exception as e:
                    print("type error: " + str(e))
                appui.exportDS("Structure Query Result", packdata)

    except Exception as e:
        log.log(repr(e))



def run_structure_query(appui, web_services, smiles, operator, similarity, username, password):
    """
    Obtain the other parameters from the user, use the web services chemical_search function to get
    the data and when it is returned package it into a dataset.

    :param: appui               The stardrop appui object
    :param: web_services        The web_services instance
    :param: smiles              The input SMILES string
    :param: operator            Whether the search should be EXACT, SIMILAR or SUBSTRUCTURE
    :param: similarity          The similarity value
    :param: username
    :param: password
    :return: dataset            The dataset of compounds found in Enamine database
    """
    try:
        cb = PythonQt.ComboBoxSelector()

        if operator in ("IN", "NOTIN"):
            cb.setWindowTitle("Set up parameters")
            cb.setText(
                "Search mode options: \nBuilding blocks are scaffolds and intermediates \nScreening compounds are for biological screening")
            options = {}
            options["Currency"] = ["USD", "EUR"]
            options["Search mode"] = ["Building blocks", "Screening compounds"]
            options["Show available amounts"] = ["Show one amount and price", "Show all available amounts and prices"]
            cb.setOptions("Currency: ", options["Currency"])
            cb.addOptions("Search mode: ", options["Search mode"])
            cb.addOptions("Show available amounts: ", options["Show available amounts"])
        else:
            cb.setWindowTitle("Set up parameters")
            cb.setText(
                "Search mode options: \nBuilding blocks are scaffolds and intermediates \nScreening compounds are for biological screening \nReadily Accessible (REAL) database is a chemical space of more than 720 Million synthetically accessible drug-like compounds")
            options = {}
            options["Currency"] = ["USD", "EUR"]
            options["Search mode"] = ["Search all", "Building blocks", "Screening compounds", "REAL database"]
            options["Show available amounts"] = ["Show one amount and price", "Show all available amounts and prices"]
            cb.setOptions("Currency: ", options["Currency"])
            cb.addOptions("Search mode: ", options["Search mode"])
            cb.addOptions("Show available amounts: ", options["Show available amounts"])
        if cb.run():  # True if user clicked OK rather than Cancel
            currency = cb.getSelectedOption(0)
            searchmode = cb.getSelectedOption(1)
            multiple_amounts = cb.getSelectedOption(2)
        try:
            dataset = None
            print("Currency: " + currency)
            print("Catalogue: " + searchmode)

            progDlg = PythonQt.SimpleProgress()
            progDlg.modal = True
            progDlg.setWindowTitle('Find Enamine Compounds')
            progDlg.setMessage('Running Enamine Web Services search')
            progDlg.show()
            progDlg.setProgress(4)
            appui.updateUI()

            # if the user searches for substructures which are really simple eg phenol or benzene, too many structures
            # are returned and API doesn't respond
            if smiles in ("C1=CC=CC=C1", "OC1=CC=CC=C1", "CC1=CC=CC=C1", "C1CCCCC1", "C1CCCC1", "C1CCNC1", "C1CCOC1",
                          "CCNC", "CC(C)=O", "OC1CCCC1", "CCC1=CC=CC=C1", "CCC1CCCCC1", "CCC1CCCC1", "NC1=CC=CC=C1",
                          "NC1CCCCC1", "NC1CCCC1", "CNC1=CC=CC=C1", "CNC1=CC=CC=C1C") and operator in ("IN", "LIKE"):
                appui.show(
                    "Too many compounds returned, please try a more specific search")
                structure_query(appui)

            else:
                progDlg.setProgress(10)
                # If the user sketched the structure and chose a similarity lower than 0.4, tell them to set this
                # higher and bring up threshold dialog.
                if operator == "LIKE":
                    if float(similarity) < 0.4:
                        appui.show("Please choose a higher similarity")
                        similarity = find_threshold(appui)

                progDlg.setProgress(15)

                molecules = web_services.chemical_search(appui,
                                                         smiles,
                                                         operator,
                                                         currency,
                                                         similarity,
                                                         searchmode)

                progDlg.setProgress(20)

                if molecules:
                    Enamine_credentials.CredentialsManager(username, password, True)
                else:
                    None

                if "data" in molecules:
                    column_definitions = [
                            ("SMILES", app.mol, None),
                            ("Enamine ID", app.string, None),
                            ("Formula", app.string, None),
                            ("Molecular Weight", app.cont, None),
                            ("CAS", app.string, None),
                            ("Catalogue URL", app.string, None),
                            ("Amount in stock (g)", app.string, None),
                            ("Amount (g)", app.cont, None),
                            ("Amount (mg)", app.cont, None),
                            ("Price (" + str(currency) + ")", app.cont, None),
                            ]

                    cols = _create_columns(column_definitions)
                    dataset = app.Dataset(cols)
                    dataset.idColName = "Enamine ID"
                    moleculelist = molecules["data"]

                    progDlg.setProgress(40)

                    if moleculelist in (None, []) and molecules["result"].get("code") in (1, None):
                        if molecules["result"].get("message") == None:
                            appui.show("No molecules found")
                        else:
                            appui.show(molecules["result"].get("message"))

                    else:
                        if moleculelist in (None, []) and molecules["result"].get("code") == 2:
                            cred.clear()
                            appui.show(
                                "Login details not recognised")
                            web_services, username, password = _create_enamine_web_services(appui)
                            molecules = web_services.chemical_search(appui,
                                                                     smiles,
                                                                     operator,
                                                                     currency,
                                                                     similarity,
                                                                     searchmode)
                            moleculelist = molecules["data"]
                            # If the user puts in the wrong username and password again, bring up error messages
                            # saying not recognised but don't prompt for user to put these in again.
                            if moleculelist in (None, []) and molecules["result"].get("code") == 2:
                                appui.show(
                                    "Login details not recognised")

                            elif moleculelist in (None, []) and molecules["result"].get("code") in (1, None):
                                if molecules["result"].get("message") == None:
                                    appui.show("No molecules found")
                                else:
                                    appui.show(
                                        molecules["result"].get("message"))
                            else:
                                None
                        else:
                            None

                        progDlg.setProgress(60)

                        if multiple_amounts == "Show all available amounts and prices":
                            dataset = multiple_amounts_for_structure_search(dataset, moleculelist, cols)

                        elif multiple_amounts == "Show one amount and price":
                            if len(moleculelist) > 0:
                                for names in moleculelist:
                                    packs = names.get("packs")
                                    if len(packs) == 0:
                                        dataset.cols[7].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
                                        dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
                                        dataset.cols[9].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
                                    else:
                                        _unpack_packs(dataset, packs)
                                    product_url = names.get("productUrl")
                                    product_link = '<a href="{0}">{0}</a>'.format(product_url)
                                    dataset.cols[0].append(
                                        app.Dataset.molecule(names.get("smile")))
                                    dataset.cols[1].append(
                                        _create_string_entry(names.get("Id")))
                                    dataset.cols[2].append(
                                        _create_string_entry(names.get("formula")))
                                    dataset.cols[3].append(
                                        app.Dataset.cont(names.get("mw"), 0.0))
                                    dataset.cols[4].append(
                                        _create_string_entry(names.get("cas")))
                                    dataset.cols[5].append(
                                        _create_string_entry(product_link))
                                    if names.get("availability") == 0.0:
                                        dataset.cols[6].append(
                                            _create_string_entry("OUT OF STOCK\n" + names.get("deliveryDays")))
                                    else:
                                        dataset.cols[6].append(
                                            _create_string_entry(names.get("availability")))

                            col_to_del = []

                            # identify any empty amount columns, list their indices and delete them
                            for i in range(len(column_definitions)):
                                valid_amount = False
                                for field in dataset.cols[i]:
                                    type_and_units = cols[i][3][0]
                                    if 'Amount' in type_and_units:
                                        if field[2][1] == 0:
                                            None
                                        else:
                                            valid_amount = True
                                    elif 'CAS' in type_and_units:
                                        if field[2] == '\x00':
                                            None
                                        else:
                                            valid_amount = True
                                    else:
                                        valid_amount = True


                                if valid_amount:
                                    None
                                else:
                                    col_to_del.append(i)

                            progDlg.setProgress(95)

                            # reverse the list of empty columns to delete them without affecting indexing
                            reverse_list = col_to_del[::-1]

                            if reverse_list:
                                for col_num in reverse_list:
                                    dataset.delColumn(col_num)
                            else:
                                None

                progDlg.setProgress(90)

                progDlg.hide()
                progDlg.delete()

        except Exception as e:
            log.log(repr(e))

        if dataset.cols[0]:
            return dataset
        else:
            return None

    except Exception as e:
        log.log(repr(e))

def selected_ids_query(appui):
    """
    Perform an ID query with ID(s) entered into a dialog box.

    One or multiple IDs can be given.

    :param appui:    The stardrop appui object
    :return: None
    """
    try:
        web_services, username, password = _create_enamine_web_services(appui)

        dlg = PythonQt.IDQueryDialog()
        if dlg.run():
            # The return value is simply the list of IDs
            ids = dlg.selectedQuery()
            log.log("running ID query %s" % (repr(ids)))

            cb = PythonQt.ComboBoxSelector()
            cb.setWindowTitle("Set up parameters")
            cb.setText("")
            options = {}
            options["Currency"] = ["USD", "EUR"]
            options["Show available amounts"] = ["Show one amount and price", "Show all available amounts and prices"]
            cb.setOptions("Currency: ", options["Currency"])
            cb.addOptions("Show available amounts: ", options["Show available amounts"])
            if cb.run():  # True if user clicked OK rather than Cancel
                currency = cb.getSelectedOption(0)
                multiple_amounts = cb.getSelectedOption(1)


        progDlg = PythonQt.SimpleProgress()
        progDlg.setWindowTitle('Find Enamine Compounds')
        progDlg.setMessage('Running Enamine Web Services search')
        progDlg.show()
        progDlg.setProgress(5)

        print(ids)

        # If the user has entered IDs separated by commas or spaces, ids will be a tuple of length one with
        # IDs separated by a comma or space so repackage these into a tuple
        for mol_id in ids:
            if len(ids) == 1:
                if ", " in mol_id:
                    ids = mol_id.split(", ")
                elif " " in mol_id:
                    ids = mol_id.split(" ")
                else:
                    None
            # If the user has entered multiple IDs on different lines AND separated by a comma, ask them
            # to enter one per line
            if len(ids) > 1:
                if ", " in mol_id:
                    appui.show("Please write one ID per line")
                    dlg = PythonQt.IDQueryDialog()
                    if dlg.run():
                        # The return value is simply the list of IDs
                        ids = dlg.selectedQuery()
                        log.log("running ID query %s" % (repr(ids)))
                else:
                    None

        if len(ids) > 0:
            try:
                data_list = []

                progDlg.setProgress(10)
                progress_start = 10
                increment = 40.0 / len(ids)

                # For every ID entered, get the data for that compound and append it to a list of data
                for mol_id in ids:
                    progDlg.setProgress(progress_start)
                    progress_start += increment
                    molecule_data = web_services.load_molecule(appui, mol_id, currency)
                    if "data" in molecule_data:
                        if molecule_data["data"] == None and molecule_data["result"].get("code") == 1:
                            appui.show(
                                "Molecule ID " + str(mol_id) + " is not recognised")
                        else:
                            if molecule_data['data'] == None and molecule_data["result"].get("code") == 2:
                                cred.clear()
                                appui.show(
                                    "Login details not recognised")
                                # Ask for login details again
                                web_services, username, password = _create_enamine_web_services(appui)
                                for mol_id in ids:
                                    molecule_data = web_services.load_molecule(appui, mol_id, currency)

                                data_list.append(molecule_data["data"])
                                if molecule_data["data"] == None:
                                    if molecule_data["result"].get("code") == 2:
                                        cred.clear()
                                        appui.show(
                                            "Login details not recognised")
                                        break
                                    elif molecule_data["result"].get("code") == 1:
                                        appui.show(
                                            "Molecule ID " + str(mol_id) + " is not recognised")
                                else:
                                    data_list.append(molecule_data["data"])
                            else:
                                data_list.append(molecule_data["data"])

                progDlg.setProgress(60)

                datalistlong = list(itertools.chain.from_iterable(data_list))
                dictdatalist = {u"data": datalistlong,  u'result': {u'message': u'Ok', u'code': 0}}
                data = dictdatalist["data"]

                column_definitions = [
                    ("SMILES", app.mol, None),
                    ("Enamine ID", app.string, None),
                    ("Formula", app.string, None),
                    ("Molecular Weight", app.cont, None),
                    ("CAS", app.string, None),
                    ("Catalogue URL", app.string, None),
                    ("Amount in stock (g)", app.string, None),
                    ("Amount", app.string, None),
                    ("Price (" + str(currency) + ")", app.cont, None),
                    ]

                cols = _create_columns(column_definitions)

                progDlg.setProgress(70)

                dataset = app.Dataset(cols)
                dataset.idColName = "Id"

                Enamine_credentials.CredentialsManager(username, password, True)

                progDlg.setProgress(75)

                if multiple_amounts == "Show all available amounts and prices":
                    multiple_amounts_for_ID_search(dataset, data)

                elif multiple_amounts == "Show one amount and price":
                    if len(data) > 0:
                        for names in data:
                            packs = names.get("packs")
                            if packs == None:
                                dataset.cols[7].append(
                                    _create_string_entry(""))
                                dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
                            else:
                                _unpack_packs_ID_search(dataset, packs)
                            product_url = names.get("productUrl")
                            product_link = '<a href="{0}">{0}</a>'.format(product_url)
                            dataset.cols[0].append(
                                app.Dataset.molecule(names.get("smile")))
                            dataset.cols[1].append(
                                _create_string_entry(names.get("Id")))
                            dataset.cols[2].append(
                                _create_string_entry(names.get("formula")))
                            dataset.cols[3].append(
                                app.Dataset.cont(names.get("mw"), 0.0))
                            dataset.cols[4].append(
                                _create_string_entry(names.get("cas")))
                            dataset.cols[5].append(
                                _create_string_entry(product_link))
                            if names.get("availability") == 0.0:
                                dataset.cols[6].append(
                                    _create_string_entry("OUT OF STOCK\n" + names.get("deliveryDays")))
                            else:
                                dataset.cols[6].append(
                                    _create_string_entry(names.get("availability")))

                col_to_del = []
                progDlg.setProgress(85)

                # identify if the CAS column is empty, if so delete it
                for i in range(len(column_definitions)):
                    valid_amount = False
                    for field in dataset.cols[i]:
                        type_and_units = cols[i][3][0]
                        if 'CAS' in type_and_units:
                            if field[2] == '\x00':
                                None
                            else:
                                valid_amount = True
                        else:
                            valid_amount = True

                    if valid_amount:
                        None
                    else:
                        col_to_del.append(i)

                for col_num in col_to_del:
                    dataset.delColumn(col_num)
                else:
                    None

                progDlg.setProgress(95)

                if dataset.cols[0]:
                    try:
                        packdata = dataset.pack()
                    except Exception as e:
                        print("type error: " + str(e))
                    appui.exportDS("ID Query Result", packdata)

            except Exception as e:
                log.log(repr(e))


    except Exception as e:
        log.log(repr(e))

def _unpack_packs_multiple(dataset, packs):

    """
    Puts the amounts and prices from packs into a dataset with one row per amount/price.

    :param dataset:         The dataset to write new rows to
    :param packs:           A list of dictionaries with amounts that are available and their prices
    :return: Dataset
    """

    for pack_ in packs:
        if pack_.get("measure") == "g":
            dataset.cols[7].append(
                app.Dataset.cont(float(pack_.get("amount")), 0.0))
        else:
            dataset.cols[7].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
        if pack_.get("measure") == "mg":
            dataset.cols[8].append(
                app.Dataset.cont(float(pack_.get("amount")), 0.0))
        else:
            dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

        dataset.cols[9].append(
            app.Dataset.cont(pack_.get("price"), 0.0))

    return dataset


def multiple_amounts_for_structure_search(dataset, moleculelist, cols):

    """
    Puts the amounts and prices from packs into a dataset with one row per amount/price. This function is called
    when the user selects to show all amounts and prices available.

    :param dataset:         The dataset to write new rows to
    :param moleculelist:    A list of data for every compound returned
    :return: Dataset
    """
    try:
        if len(moleculelist) > 0:
            for names in moleculelist:
                packs = names.get("packs")
                if len(packs) == 0:
                    product_url = names.get("productUrl")
                    product_link = '<a href="{0}">{0}</a>'.format(product_url)
                    dataset.cols[0].append(
                        app.Dataset.molecule(names.get("smile")))
                    dataset.cols[1].append(
                        _create_string_entry(names.get("Id")))
                    dataset.cols[2].append(
                        _create_string_entry(names.get("formula")))
                    dataset.cols[3].append(
                        app.Dataset.cont(names.get("mw"), 0.0))
                    dataset.cols[4].append(
                        _create_string_entry(names.get("cas")))
                    dataset.cols[5].append(
                        _create_string_entry(product_link))
                    if names.get("availability") == 0.0:
                        dataset.cols[6].append(
                            _create_string_entry("OUT OF STOCK\n" + names.get("deliveryDays")))
                    else:
                        dataset.cols[6].append(
                            _create_string_entry(names.get("availability")))
                    dataset.cols[7].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
                    dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))
                    dataset.cols[9].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

                elif len(packs) > 0:
                    """Going to unpack the packs object which has information on the different amounts and prices for this compound"""
                    _unpack_packs_multiple(dataset, packs)
                    product_url = names.get("productUrl")
                    product_link = '<a href="{0}">{0}</a>'.format(product_url)
                    for i in range(len(packs)):
                        dataset.cols[0].append(
                            app.Dataset.molecule(names.get("smile")))
                        dataset.cols[1].append(
                            _create_string_entry(names.get("Id")))
                        dataset.cols[2].append(
                            _create_string_entry(names.get("formula")))
                        dataset.cols[3].append(
                            app.Dataset.cont(names.get("mw"), 0.0))
                        dataset.cols[4].append(
                            _create_string_entry(names.get("cas")))
                        dataset.cols[5].append(
                            _create_string_entry(product_link))
                        if names.get("availability") == 0.0:
                            dataset.cols[6].append(
                                _create_string_entry("OUT OF STOCK\n" + names.get("deliveryDays")))
                        else:
                            dataset.cols[6].append(
                                _create_string_entry(names.get("availability")))
            col_to_del = []

            # Identify whether an amount column or the CAS column is empty, delete if empty
            for i in range(9):
                valid_amount = False
                for field in dataset.cols[i]:
                    type_and_units = cols[i][3][0]
                    if 'Amount' in type_and_units:
                        if field[2][1] == 0:
                            None
                        else:
                            valid_amount = True
                    elif 'CAS' in type_and_units:
                        if field[2] == '\x00':
                            None
                        else:
                            valid_amount = True
                    else:
                        valid_amount = True

                if valid_amount:
                    None
                else:
                    col_to_del.append(i)

            # reverse the list of empty columns to delete them without affecting indexing
            reverse_list = col_to_del[::-1]

            if reverse_list:
                for col_num in reverse_list:
                    dataset.delColumn(col_num)
            else:
                None

            return dataset

        else:
            appui.show("No molecules found")

    except Exception as e:
        log.log(repr(e))


def _unpack_packs_multiple_ID_search(dataset, packs):

    """
    Put the amounts and prices in packs into the dataset with one row per amount/price. This function is called
    when the user selects to show all amounts and prices available.

    :param dataset:         The dataset to write new rows to
    :param packs:           A list of dictionaries with amounts that are available and their prices
    :return: Dataset
    """

    for pack_ in packs:
        dataset.cols[7].append(
            _create_string_entry(str(pack_.get("amount")) + " " + str(pack_.get("measure"))))
        dataset.cols[8].append(
            app.Dataset.cont(pack_.get("price"), 0.0))

    return dataset


def multiple_amounts_for_ID_search(dataset, data):

    """
    Put the amounts and prices in packs into the dataset with one row per amount/price. This function is called
    when the user selects to show all amounts and prices available.

    :param dataset:         The dataset to write new rows to
    :param data:            A list of data for every compound returned
    :return: Dataset
    """
    try:
        if len(data) > 0:
            for names in data:
                packs = names.get("packs")
                if len(packs) == 0:
                    product_url = names.get("productUrl")
                    product_link = '<a href="{0}">{0}</a>'.format(product_url)
                    dataset.cols[0].append(
                        app.Dataset.molecule(names.get("smile")))
                    dataset.cols[1].append(
                        _create_string_entry(names.get("Id")))
                    dataset.cols[2].append(
                        _create_string_entry(names.get("formula")))
                    dataset.cols[3].append(
                        app.Dataset.cont(names.get("mw"), 0.0))
                    dataset.cols[4].append(
                        _create_string_entry(names.get("cas")))
                    dataset.cols[5].append(
                        _create_string_entry(product_link))
                    if names.get("availability") == 0.0:
                        dataset.cols[6].append(
                            _create_string_entry("OUT OF STOCK\n" + names.get("deliveryDays")))
                    else:
                        dataset.cols[6].append(
                            _create_string_entry(names.get("availability")))
                    dataset.cols[7].append(
                        _create_string_entry(""))
                    dataset.cols[8].append((12 + 2 * 8 + 1, (False, [], []), ('\x00', 0, 0.0)))

                elif len(packs) > 0:
                    """Going to unpack packs which has information on the different amounts and prices for this compound"""
                    _unpack_packs_multiple_ID_search(dataset, packs)
                    product_url = names.get("productUrl")
                    product_link = '<a href="{0}">{0}</a>'.format(product_url)
                    for i in range(len(packs)):
                        dataset.cols[0].append(
                            app.Dataset.molecule(names.get("smile")))
                        dataset.cols[1].append(
                            _create_string_entry(names.get("Id")))
                        dataset.cols[2].append(
                            _create_string_entry(names.get("formula")))
                        dataset.cols[3].append(
                            app.Dataset.cont(names.get("mw"), 0.0))
                        dataset.cols[4].append(
                            _create_string_entry(names.get("cas")))
                        dataset.cols[5].append(
                            _create_string_entry(product_link))
                        if names.get("availability") == 0.0:
                            dataset.cols[6].append(
                                _create_string_entry("OUT OF STOCK\n" + names.get("deliveryDays")))
                        else:
                            dataset.cols[6].append(
                                _create_string_entry(names.get("availability")))

            return dataset

        else:
            print("No molecules found")

    except Exception as e:
        log.log(repr(e))

                
def get_stardrop_definitions():
    definitions = [
        {
            'script_name': 'Enamine/Sketch structure to use in search...',
            'callback': sketch_structure_query
        },
        {
            'script_name': 'Enamine/Search by similarity to selected molecule...',
            'callback': structure_query_selected
        },
        {
            'script_name': 'Enamine/Search by Enamine ID...',
            'callback': selected_ids_query
        }
    ]
    
    return definitions
